import logging
from typing import Callable

import confluent_kafka  # type: ignore

logger = logging.getLogger("adc-streaming")


ErrorCallback = Callable[[confluent_kafka.KafkaError], None]


def log_client_errors(kafka_error: confluent_kafka.KafkaError):
    if kafka_error.code() == confluent_kafka.KafkaError._ALL_BROKERS_DOWN:
        # This error occurs very frequently. It's not nearly as fatal as it
        # sounds: it really indicates that the client's broker metadata has
        # timed out. It appears to get triggered in races during client
        # shutdown, too. See https://github.com/edenhill/librdkafka/issues/2543
        # for more background.
        logger.warn("client is currently disconnected from all brokers")
    else:
        logger.error(f"internal kafka error: {kafka_error}")
        raise(KafkaException.from_kafka_error(kafka_error))


DeliveryCallback = Callable[[confluent_kafka.KafkaError, confluent_kafka.Message], None]


def log_delivery_errors(
        kafka_error: confluent_kafka.KafkaError,
        msg: confluent_kafka.Message) -> None:
    if kafka_error is not None:
        logger.error(f"delivery error: {kafka_error}")


def raise_delivery_errors(kafka_error: confluent_kafka.KafkaError,
                          msg: confluent_kafka.Message) -> None:
    if kafka_error is not None:
        raise KafkaException.from_kafka_error(kafka_error)
    elif msg.error() is not None:
        raise KafkaException.from_kafka_error(msg.error())


class KafkaException(Exception):
    @classmethod
    def from_kafka_error(cls, error):
        return cls(error)

    def __init__(self, error):
        self.error = error
        self.name = error.name()
        self.reason = error.str()
        self.retriable = error.retriable()
        self.fatal = error.fatal()
        msg = f"Error communicating with Kafka: code={self.name} {self.reason}"
        super(KafkaException, self).__init__(msg)



import abc
import dataclasses
import logging
from datetime import timedelta
from typing import Dict, List, Optional, Union

import confluent_kafka  # type: ignore

from .auth import SASLAuth
from .errors import (DeliveryCallback, ErrorCallback, log_client_errors,
                     log_delivery_errors)


class Producer:
    conf: 'ProducerConfig'
    _producer: confluent_kafka.Producer
    logger: logging.Logger

    def __init__(self, conf: 'ProducerConfig') -> None:
        self.logger = logging.getLogger("adc-streaming.producer")
        self.conf = conf
        self.logger.debug(f"connecting to producer with config {conf._to_confluent_kafka()}")
        self._producer = confluent_kafka.Producer(conf._to_confluent_kafka())

    def write(self,
              msg: Union[bytes, 'Serializable'],
              headers: Optional[Union[dict, list]] = None,
              delivery_callback: Optional[DeliveryCallback] = log_delivery_errors) -> None:
        if isinstance(msg, Serializable):
            msg = msg.serialize()
        self.logger.debug("writing message to %s", self.conf.topic)
        if delivery_callback is not None:
            self._producer.produce(self.conf.topic, msg, headers=headers,
                                   on_delivery=delivery_callback)
        else:
            self._producer.produce(self.conf.topic, msg, headers=headers)

    def flush(self, timeout: timedelta = timedelta(seconds=10)) -> int:
        """Attempt to flush enqueued messages. Return the number of messages still
        enqueued after the attempt.

        """
        n = self._producer.flush(timeout.total_seconds())
        if n > 0:
            self.logger.debug("flushed messages, %d still enqueued", n)
        else:
            self.logger.debug("flushed all messages")
        return n

    def close(self) -> int:
        self.logger.debug("shutting down producer")
        return self.flush()

    def __enter__(self) -> 'Producer':
        return self

    def __exit__(self, type, value, traceback) -> bool:
        if type == KeyboardInterrupt:
            print("Aborted (CTRL-C).")
            return True
        if type is None and value is None and traceback is None:
            n_unsent = self.close()
            if n_unsent > 0:
                raise Exception(f"{n_unsent} messages remain unsent, some data may have been lost!")
            return False
        return False


@dataclasses.dataclass
class ProducerConfig:
    broker_urls: List[str]
    topic: str
    auth: Optional[SASLAuth] = None
    error_callback: Optional[ErrorCallback] = log_client_errors

    # produce_timeout sets the maximum amount of time that the backend can take
    # to send a message to Kafka. Use a value of 0 to never timeout.
    produce_timeout: timedelta = timedelta(seconds=10)

    # produce_backoff_time sets the time the backend will wait before retrying
    # to send a message to Kafka. May not be less than one millisecond.
    produce_backoff_time: timedelta = timedelta(milliseconds=100)

    # use_idempotence instructs the backend whether to ensure that messages are
    # recorded by the broker exactly once and in the order of production.
    use_idempotence: bool = False

    # reconnect_backoff_time is the time that the backend should initially wait
    # before attempting to reconnect to Kafka if its connection fails.
    # Repeated failures will cause the wait time to be increased exponentially,
    # with a random variation, until reconnect_max_time is reached.
    reconnect_backoff_time: timedelta = timedelta(milliseconds=100)

    # reconnect_max_time is the longest time that the backend should wait
    # between attempts to reconnect to Kafka.
    reconnect_max_time: timedelta = timedelta(seconds=10)

    def _to_confluent_kafka(self) -> Dict:
        def as_ms(td: timedelta):
            """Convert a timedelta object to a duration in milliseconds"""
            return int(td.total_seconds() * 1000.0)

        if self.produce_backoff_time < timedelta(milliseconds=1):
            raise ValueError("produce_backoff_time may not be less than one millisecond")
        config = {
            "bootstrap.servers": ",".join(self.broker_urls),
            "enable.idempotence": self.use_idempotence,
            "message.timeout.ms": as_ms(self.produce_timeout),
            "reconnect.backoff.max.ms": as_ms(self.reconnect_max_time),
            "reconnect.backoff.ms": as_ms(self.reconnect_backoff_time),
            "retry.backoff.ms": as_ms(self.produce_backoff_time),
        }
        if self.error_callback is not None:
            config["error_cb"] = self.error_callback
        if self.auth is not None:
            config.update(self.auth())
        return config


class Serializable(abc.ABC):
    def serialize(self) -> bytes:
        raise NotImplementedError()



import logging
import tempfile
import time
import unittest
from datetime import timedelta
from typing import List

import docker
import pytest

import adc.consumer
import adc.io
import adc.producer

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("adc-streaming").setLevel(logging.DEBUG)
logger = logging.getLogger("adc-streaming.tests")


@pytest.mark.integration_test
class KafkaIntegrationTestCase(unittest.TestCase):
    """This test runs a Kafka broker in a Docker container, and makes sure that
    messages can round-trip through that broker, testing both the Producer and
    Consumer sides of the library.

    It can be pretty slow, since it has to set up a container and wait for Kafka
    to come online.

    """
    @classmethod
    def setUpClass(cls):
        cls.kafka = KafkaDockerConnection()

    @classmethod
    def tearDownClass(cls):
        cls.kafka.close()

    def test_round_trip(self):
        """Try writing a message into the Kafka broker, and try pulling the same
        message back out.

        """
        topic = "test_round_trip"
        # Push one message in...
        simple_write_msg(self.kafka, topic, "can you hear me?")
        # ... and pull it back out.
        consumer = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer",
            auth=self.kafka.auth,
        ))
        consumer.subscribe(topic)
        stream = consumer.stream()

        msg = next(stream)
        if msg.error() is not None:
            raise Exception(msg.error())

        self.assertEqual(msg.topic(), topic)
        self.assertEqual(msg.value(), b"can you hear me?")

    @unittest.skip("skipping due to bug in librdkafka")
    def test_consume_from_end(self):
        # Write a few messages.
        topic = "test_consume_from_end"
        simple_write_msgs(self.kafka, topic, [
            "message 1",
            "message 2",
            "message 3",
        ])
        # Start a consumer from the end position
        consumer = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer",
            auth=self.kafka.auth,
            start_at=adc.consumer.ConsumerStartPosition.LATEST,
        ))
        consumer.subscribe(topic)
        stream = consumer.stream()

        # Now add messages after the "end"
        simple_write_msg(self.kafka, topic, "message 4")

        msg = next(stream)
        self.assertEqual(msg.topic(), topic)
        self.assertEqual(msg.value(), b"message 4")

    def test_consume_from_beginning(self):
        # Write a few messages.
        topic = "test_consume_from_beginning"
        batch = [
            "message 1",
            "message 2",
            "message 3",
            "message 4",
        ]
        simple_write_msgs(self.kafka, topic, batch)

        # Start a consumer from the beginning.
        consumer = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer",
            auth=self.kafka.auth,
            read_forever=False,
            start_at=adc.consumer.ConsumerStartPosition.EARLIEST,
        ))
        consumer.subscribe(topic)
        stream = consumer.stream()
        msgs = [msg for msg in stream]

        self.assertEqual(len(batch), len(msgs))
        for expected, actual in zip(batch, msgs):
            self.assertEqual(actual.topic(), topic)
            self.assertEqual(actual.value().decode(), expected)

    def test_consume_stored_offsets(self):
        # Write first batch of messages.
        topic = "test_stored_offsets"
        batch_1 = [
            "message 1",
            "message 2",
            "message 3",
            "message 4",
        ]
        simple_write_msgs(self.kafka, topic, batch_1)

        # Start first consumer, reading from earliest offset.
        consumer_1 = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer_1",
            auth=self.kafka.auth,
            read_forever=False,
            start_at=adc.consumer.ConsumerStartPosition.EARLIEST,
        ))
        consumer_1.subscribe(topic)
        stream_1 = consumer_1.stream()
        msgs_1 = [msg for msg in stream_1]

        # Check that all messages from first batch are processed.
        self.assertEqual(len(batch_1), len(msgs_1))
        for expected, actual in zip(batch_1, msgs_1):
            self.assertEqual(actual.topic(), topic)
            self.assertEqual(actual.value().decode(), expected)

        # Write second batch of messages.
        batch_2 = [
            "message 5",
            "message 6",
            "message 7",
        ]
        simple_write_msgs(self.kafka, topic, batch_2)

        # Read more messages from first consumer. This should now
        # only read from the stored offset, so that only the second
        # batch is processed.
        stream_1 = consumer_1.stream()
        msgs_1 = [msg for msg in stream_1]
        self.assertEqual(len(batch_2), len(msgs_1))
        for expected, actual in zip(batch_2, msgs_1):
            self.assertEqual(actual.topic(), topic)
            self.assertEqual(actual.value().decode(), expected)

        # Start second consumer, also reading from earliest offset.
        consumer_2 = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer_2",
            auth=self.kafka.auth,
            read_forever=False,
            start_at=adc.consumer.ConsumerStartPosition.EARLIEST,
        ))
        consumer_2.subscribe(topic)
        stream_2 = consumer_2.stream()
        msgs_2 = [msg for msg in stream_2]

        # Now check that messages from both batches are processed.
        self.assertEqual(len(batch_1 + batch_2), len(msgs_2))
        for expected, actual in zip(batch_1 + batch_2, msgs_2):
            self.assertEqual(actual.topic(), topic)
            self.assertEqual(actual.value().decode(), expected)

    def test_consume_not_forever(self):
        topic = "test_consume_not_forever"
        simple_write_msg(self.kafka, topic, "message 1")

        consumer = adc.consumer.Consumer(adc.consumer.ConsumerConfig(
            broker_urls=[self.kafka.address],
            group_id="test_consumer",
            auth=self.kafka.auth,
            read_forever=False
        ))
        consumer.subscribe(topic)
        stream = consumer.stream()

        msg = next(stream)
        if msg.error() is not None:
            raise Exception(msg.error())
        self.assertEqual(msg.topic(), topic)
        self.assertEqual(msg.value(), b"message 1")
        with self.assertRaises(StopIteration):
            next(stream)

    def test_contextmanager_support(self):
        topic = "test_contextmanager_support"
        url = f"kafka://{self.kafka.address}/{topic}"
        with adc.io.open(url, mode="w", auth=self.kafka.auth) as p:
            p.write("message 1")
            p.write("message 2")
            p.write("message 3")

        logger.info("done with writes")

        group = "test_contextmanager_group"
        url = f"kafka://{group}@{self.kafka.address}/{topic}"
        with adc.io.open(url, mode="r", auth=self.kafka.auth, read_forever=False) as stream:
            messages = [m for m in stream]
        logger.info("done with reads")
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0].value(), b"message 1")
        self.assertEqual(messages[1].value(), b"message 2")
        self.assertEqual(messages[2].value(), b"message 3")


class KafkaDockerConnection:
    """Holds connection information for communicating with a Kafka broker running
    inside a docker container.

    """

    def __init__(self):
        """Starts a Docker container running a Kafka broker, waits for it to come
        online, and prepares authentication credentials for connecting to the
        broker.

        """
        self.docker_client = docker.from_env()

        logger.info("setting up network")
        self.net = self.get_or_create_docker_network()
        logger.info("setting up container")
        self.container = self.get_or_create_container()
        logger.info("getting kafka address")
        self.address = self.poll_for_kafka_broker_address()

        logger.info("waiting for kafka to come online")
        self.poll_for_kafka_active()

        logger.info("setting up auth")
        self.certfile = tempfile.NamedTemporaryFile(
            prefix="adc-integration-test-",
            suffix=".pem",
            mode="w+b",
        )
        self.certfile.write(self.get_broker_cert(self.container))
        self.certfile.flush()
        logger.info(f"certfile written to {self.certfile.name}")
        self.auth = adc.auth.SASLAuth(
            user="test", password="test-pass",
            ssl_ca_location=self.certfile.name,
        )

    def poll_for_kafka_broker_address(self, maxiter=20, sleep=timedelta(milliseconds=500)):
        """Block until the Docker daemon tells us the IP and Port of the Kafa broker.

        Returns the ip and port as a string in the form "ip:port."
"""
        i = 0
        while (not self.query_kafka_broker_address()) and i < maxiter:
            logger.info("polling to wait for container to acquire port...")
            time.sleep(sleep.total_seconds())
            i = i + 1
            self.container.reload()
        assert i < maxiter
        return self.query_kafka_broker_address()

    def query_kafka_broker_address(self):
        """Ask the Docker API for the exposed port of the Kafka broker."""
        addrs = self.container.attrs.get("NetworkSettings", {}).get(
            "Ports", {}).get("9092/tcp", [])
        if not addrs:
            return None
        ip = addrs[0]['HostIp']
        port = addrs[0]['HostPort']
        return f"{ip}:{port}"

    def poll_for_kafka_active(self, maxiter=20, sleep=timedelta(milliseconds=500)):
        """Block until Kafka's network listener is accepting connections."""
        i = 0
        while (not self.query_kafka_active()) and i < maxiter:
            logger.info("polling to wait for kafka to come online...")
            time.sleep(sleep.total_seconds())
            i = i + 1
        assert i < maxiter

    def query_kafka_active(self):
        """Returns True if the Kafka broker's listener is accepting connections.

        This works by running netcat within the container and checking its exit code.
        """
        exit_code, _ = self.container.exec_run(
            "/bin/nc localhost 9092",
        )
        return exit_code == 0

    def get_broker_cert(self, container):
        """Returns the byte string contents of the generated TLS certificate
        used by the broker.
        """
        code, output = container.exec_run(
            "/bin/cat /root/shared/tls/cacert.pem")
        if code != 0:
            raise AssertionError(b"failed to get broker cert:" + output)
        return output

    def close(self):
        """Closes open files, shuts down containers, and tears down the docker network.

        """
        self.certfile.close()
        logger.info("tearing down container")
        self.container.stop()
        logger.info("tearing down network")
        self.net.remove()

    def exec_in_container(self, cmd):
        exit_code, output = self.container.exec_run(cmd)
        msg = f"Exit error running {cmd}: {output}"
        self.assertEqual(exit_code, 0, msg)
        return output

    def get_or_create_container(self):
        """Starts a scimma/server container named 'adc-integration-test-server' and
        returns a handle referencing the container. If a container with that
        name is already running, it's returned instead.
        """
        containers = self.docker_client.containers.list(
            filters={"name": "adc-integration-test-server"},
        )
        if containers:
            return containers[0]

        return self.docker_client.containers.run(
            image="scimma/server:latest",
            name="adc-integration-test-server",
            detach=True,
            auto_remove=True,
            network=self.net.name,
            # Setting None below the OS pick an ephemeral port.
            ports={"9092/tcp": None},
        )

    def get_or_create_docker_network(self):
        """Returns a docker network named adc-integration-test, creating it if it
        doesn't exist already.

        """
        nets = self.docker_client.networks.list(
            names="adc-integration-test")
        if nets:
            return nets[0]
        return self.docker_client.networks.create(name="adc-integration-test")


def simple_write_msg(conn: KafkaDockerConnection, topic: str, msg: str):
    producer = adc.producer.Producer(adc.producer.ProducerConfig(
        broker_urls=[conn.address],
        topic=topic,
        auth=conn.auth,
    ))
    producer.write(msg)
    producer.flush()


def simple_write_msgs(conn: KafkaDockerConnection, topic: str, msgs: List[str]):
    producer = adc.producer.Producer(adc.producer.ProducerConfig(
        broker_urls=[conn.address],
        topic=topic,
        auth=conn.auth,
    ))
    for m in msgs:
        producer.write(m)
    producer.flush()



#!/usr/bin/env python

from enum import Enum

import certifi


class SASLMethod(Enum):
    """SASL method to use for authentication.
    """

    PLAIN = 1
    SCRAM_SHA_256 = 2
    SCRAM_SHA_512 = 3

    def __str__(self):
        return self.name.replace("_", "-")


class SASLAuth(object):
    """Attach SASL-based authentication to a client.

    Returns client-based auth options when called.

    Parameters
    ----------
    user : `str`
        Username to authenticate with.
    password : `str`
        Password to authenticate with.
    ssl : `bool`, optional
        Whether to enable SSL (enabled by default).
    method : `SASLMethod`, optional
        The SASL method to authenticate, default = SASLMethod.PLAIN.
        See valid SASL methods in SASLMethod.
    ssl_ca_location : `str`, optional
        If using SSL via a self-signed cert, a path/location
        to the certificate.

    """

    def __init__(self, user, password, ssl=True, method=SASLMethod.PLAIN, **kwargs):
        self._method = method

        # set up SSL options
        if ssl:
            if "ssl_ca_location" in kwargs:
                ssl_cert = kwargs["ssl_ca_location"]
            else:
                ssl_cert = certifi.where()

            self._config = {
                "security.protocol": "SASL_SSL",
                "ssl.ca.location": ssl_cert,
            }
        else:
            self._config = {"security.protocol": "SASL_PLAINTEXT"}

        # set up SASL options
        self._config["sasl.mechanism"] = str(self._method)
        self._config["sasl.username"] = user
        self._config["sasl.password"] = password

    def __call__(self):
        return self._config



import logging
import warnings
from contextlib import contextmanager
from typing import Iterable, List, Optional, Union

import confluent_kafka  # type: ignore

from adc import auth, consumer, kafka, producer

logger = logging.getLogger("adc-streaming")


def open(url: str,
         mode: str = 'r',
         auth: Optional[auth.SASLAuth] = None,
         start_at: consumer.ConsumerStartPosition = consumer.ConsumerStartPosition.EARLIEST,  # noqa: E501
         read_forever: bool = True,
         ) -> Union[producer.Producer, Iterable[confluent_kafka.Message]]:
    group_id, broker_addresses, topics = kafka.parse_kafka_url(url)
    logger.debug("connecting to addresses=%s  group_id=%s  topics=%s",
                 broker_addresses, group_id, topics)
    if mode == "r":
        if group_id is None:
            raise ValueError("group ID must be set when in reader mode")
        return _open_consumer(group_id, broker_addresses, topics, auth, start_at, read_forever)
    elif mode == "w":
        if len(topics) != 1:
            raise ValueError("must specify exactly one topic in write mode")
        if group_id is not None:
            warnings.warn("group ID has no effect when opening a stream in write mode")
        if start_at is not consumer.ConsumerStartPosition.EARLIEST:
            warnings.warn("start_at has no effect when opening a stream in write mode")
        if read_forever is not True:
            warnings.warn("read_forever has no effect when opening a stream in write mode")
        return _open_producer(broker_addresses, topics[0], auth)
    else:
        raise ValueError("mode must be either 'w' or 'r'")


@contextmanager
def _open_consumer(
        group_id: str,
        broker_addresses: List[str],
        topics: List[str],
        auth: Optional[auth.SASLAuth],
        start_at: consumer.ConsumerStartPosition,
        read_forever: bool,
) -> Iterable[confluent_kafka.Message]:
    client = consumer.Consumer(consumer.ConsumerConfig(
        broker_urls=broker_addresses,
        group_id=group_id,
        auth=auth,
        start_at=start_at,
        read_forever=read_forever,
    ))
    for t in topics:
        client.subscribe(t)

    try:
        yield client.stream()
    finally:
        client.close()


def _open_producer(
        broker_addresses: List[str],
        topic: str,
        auth: Optional[auth.SASLAuth],
) -> producer.Producer:
    return producer.Producer(producer.ProducerConfig(
        broker_urls=broker_addresses,
        topic=topic,
        auth=auth,
    ))



from urllib.parse import urlparse


def parse_kafka_url(val):
    """Extracts the group ID, broker addresses, and topic names from a Kafka URL.

    The URL should be in this form:
        ``kafka://[groupid@]broker[,broker2[,...]]/topic[,topic2[,...]]``

    The returned group ID and topic may be None if they aren't in the URL.

    """
    parsed = urlparse(val)
    if parsed.scheme != "kafka":
        raise ValueError("invalid kafka URL: must start with 'kafka://'")

    split_netloc = parsed.netloc.split("@", maxsplit=1)
    if len(split_netloc) == 2:
        group_id = split_netloc[0]
        broker_addresses = split_netloc[1].split(",")
    else:
        group_id = None
        broker_addresses = split_netloc[0].split(",")

    topics = parsed.path.lstrip("/")
    if len(topics) == 0:
        split_topics = None
    else:
        split_topics = topics.split(",")
    return group_id, broker_addresses, split_topics



try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    # NOTE: remove after dropping support for Python < 3.8
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("adc-streaming")
except PackageNotFoundError:
    # package is not installed
    pass



import dataclasses
import enum
import logging
from datetime import timedelta
from typing import Dict, Iterable, Iterator, List, Optional, Set, Union

import confluent_kafka  # type: ignore
import confluent_kafka.admin  # type: ignore

from .auth import SASLAuth
from .errors import ErrorCallback, log_client_errors


class Consumer:
    conf: 'ConsumerConfig'
    _consumer: confluent_kafka.Consumer
    logger: logging.Logger

    def __init__(self, conf: 'ConsumerConfig') -> None:
        self.logger = logging.getLogger("adc-streaming.consumer")
        self.conf = conf
        self._consumer = confluent_kafka.Consumer(conf._to_confluent_kafka())

    def subscribe(self,
                  topics: Union[str, Iterable],
                  timeout: timedelta = timedelta(seconds=10)):
        """Subscribes to topics for consuming. This method doesn't use Kafka's
        Consumer Groups; it assigns all partitions manually to this
        process.

        The topics must already exist for the subscription to succeed.
        """
        if isinstance(topics, str):
            topics = [topics]

        assignment = []
        for topic in topics:
            self.logger.debug(f"subscribing to topic {topic}")

            try:
                topic_meta = self.describe_topic(topic, timeout)
            except KeyError:
                raise ValueError(f"topic {topic} does not exist on the broker, so can't subscribe")

            for partition_id in topic_meta.partitions.keys():
                self.logger.debug(f"adding subscription to topic partition={partition_id}")
                tp = confluent_kafka.TopicPartition(
                    topic=topic,
                    partition=partition_id,
                )
                assignment.append(tp)

        self.logger.debug("registering topic assignment")
        self._consumer.assign(assignment)

    def describe_topic(
            self,
            topic: str,
            timeout: timedelta = timedelta(seconds=5.0)) -> confluent_kafka.admin.TopicMetadata:
        """Fetch confluent_kafka.admin.TopicMetadata describing a topic.
        """
        self.logger.debug(f"fetching cluster metadata to describe topic name={topic}")
        cluster_meta = self._consumer.list_topics(timeout=timeout.total_seconds())
        self.logger.debug(f"cluster metadata: {cluster_meta.topics}")
        return cluster_meta.topics[topic]

    def mark_done(self, msg: confluent_kafka.Message, asynchronous: bool = True):
        """
        Mark a message as fully-processed. In the background, the client will
        continuously synchronize this information with Kafka so that the stream can be
        resumed from this point in the future.

        If asynchronous is set to False, however, the information will be sent
        to Kafka immediately. This option allows fine-grained reliablity, but
        can seriously reduce throughput if used on every message.
        """
        if asynchronous:
            self._consumer.store_offsets(msg)
        else:
            self._consumer.commit(msg, asynchronous=False)

    def stream(self,
               autocommit: bool = True,
               batch_size: int = 100,
               batch_timeout: timedelta = timedelta(seconds=1.0)
               ) -> Iterator[confluent_kafka.Message]:
        """Returns a stream which iterates over the messages in the topics
        to which the client is subscribed.

        If autocommit is true, then messages are automatically marked as handled
        when they are yielded. This removes the need to call
        'mark_done' on each message. Callers using asynchronous message
        processing or with complex processing needs should disable this.

        batch_size controls the number of messages to request from Kafka per
        batch. Higher values may be more efficient, but may add latency.

        batch_timeout controls how long the client should wait for Kafka to
        provide a full batch of batch_size messages. Higher values may be more
        efficient, but may add latency.

        If the consumer's configuration has read_forever set to False, then the
        stream stops when the client has hit the last message in all partitions.
        This set of partitions is calculated just once when iterate() is first
        called; calling subscribe() after iterate() may cause inconsistent
        behavior in this case.

        """
        if self.conf.read_forever:
            return self._stream_forever(autocommit, batch_size, batch_timeout)
        else:
            return self._stream_until_eof(autocommit, batch_size, batch_timeout)

    def _stream_forever(self,
                        autocommit: bool = True,
                        batch_size: int = 100,
                        batch_timeout: timedelta = timedelta(seconds=1.0),
                        ) -> Iterator[confluent_kafka.Message]:
        while True:
            messages = self._consumer.consume(batch_size, batch_timeout.total_seconds())
            try:
                for m in messages:
                    err = m.error()
                    if err is None:
                        self.logger.debug(f"read message from partition {m.partition()}")
                        # Automatically mark message as processed, if desired
                        if autocommit:
                            self.mark_done(m, asynchronous=True)
                        yield m
                    else:
                        raise(confluent_kafka.KafkaException(err))
            finally:
                if autocommit:
                    self._consumer.commit(asynchronous=True)

    def _stream_until_eof(self,
                          autocommit: bool = True,
                          batch_size: int = 100,
                          batch_timeout: timedelta = timedelta(seconds=1.0),
                          ) -> Iterator[confluent_kafka.Message]:
        assignment = self._consumer.assignment()

        # Make a map of topic-name -> set of partition IDs we're assigned to.
        # When we hit a partition EOF, remove that partition from the map.
        active_partitions: Dict[str, Set[int]] = {}
        for tp in assignment:
            if tp.topic not in active_partitions:
                active_partitions[tp.topic] = set()
            self.logger.debug(f"tracking until eof for topic={tp.topic} partition={tp.partition}")
            active_partitions[tp.topic].add(tp.partition)

        while len(active_partitions) > 0:
            messages = self._consumer.consume(batch_size, batch_timeout.total_seconds())
            try:
                for m in messages:
                    err = m.error()
                    if err is None:
                        self.logger.debug(f"read message from partition {m.partition()}")
                        # Automatically mark message as processed, if desired
                        if autocommit:
                            self.mark_done(m, asynchronous=True)
                        yield m
                    elif err.code() == confluent_kafka.KafkaError._PARTITION_EOF:
                        self.logger.debug(f"eof for topic={m.topic()} partition={m.partition()}")
                        # Done with this partition, remove it
                        partition_set = active_partitions[m.topic()]
                        partition_set.remove(m.partition())
                        if len(partition_set) == 0:
                            # Done with all partitions for the topic, remove it
                            del active_partitions[m.topic()]
                    else:
                        raise(confluent_kafka.KafkaException(err))
            finally:
                if autocommit:
                    self._consumer.commit(asynchronous=True)

    def close(self):
        """ Close the consumer, ending its subscriptions. """
        self._consumer.close()


class ConsumerStartPosition(enum.Enum):
    EARLIEST = 1
    LATEST = 2

    def __str__(self):
        return self.name.lower()


@dataclasses.dataclass
class ConsumerConfig:
    broker_urls: List[str]
    group_id: str

    # When we have reached the last message on a topic, should we hold the
    # stream open to wait for more messages?
    read_forever: bool = True

    # When reading a topic for the first time, where should we start in the
    # stream? Note that, if the topic has already been consumed under the
    # provided group_id, then consumption will start after the last message that
    # was marked done with consumer.mark_done, regardless of this setting. This
    # is only used when the position in the stream is unknown.
    #
    # This is specified as a logical offset via a ConsumerStartPosition value.
    start_at: ConsumerStartPosition = ConsumerStartPosition.EARLIEST

    # Authentication package to pass in to read from Kafka.
    auth: Optional[SASLAuth] = None

    # Callback to execute whenever an internal Kafka error occurs.
    error_callback: Optional[ErrorCallback] = log_client_errors

    # How often should we save our progress to Kafka?
    offset_commit_interval: timedelta = timedelta(seconds=5)

    # Whether ncoming message CRCs should be checked to detect corruption in
    # transit. Enabling this option has a small CPU use/throughput cost.
    check_crcs: bool = False

    # reconnect_backoff_time is the time that the backend should initially wait
    # before attempting to reconnect to Kafka if its connection fails.
    # Repeated failures will cause the wait time to be increased exponentially,
    # with a random variation, until reconnect_max_time is reached.
    reconnect_backoff_time: timedelta = timedelta(milliseconds=100)

    # reconnect_max_time is the longest time that the backend should wait
    # between attempts to reconnect to Kafka.
    reconnect_max_time: timedelta = timedelta(seconds=10)

    def _to_confluent_kafka(self) -> Dict:
        def as_ms(td: timedelta):
            """Convert a timedelta object to a duration in milliseconds"""
            return int(td.total_seconds() * 1000.0)

        config = {
            "bootstrap.servers": ",".join(self.broker_urls),
            "check.crcs": self.check_crcs,
            "error_cb": self.error_callback,
            "group.id": self.group_id,
            "enable.auto.commit": True,
            "auto.commit.interval.ms": as_ms(self.offset_commit_interval),
            "enable.auto.offset.store": False,
            "queued.min.messages": 1000,
            "enable.partition.eof": not self.read_forever,
            "reconnect.backoff.max.ms": as_ms(self.reconnect_max_time),
            "reconnect.backoff.ms": as_ms(self.reconnect_backoff_time),
        }
        if self.start_at is ConsumerStartPosition.EARLIEST:
            default_topic_config = config.get("default.topic.config", {})
            default_topic_config = {
                "auto.offset.reset": "EARLIEST",
            }
            config["default.topic.config"] = default_topic_config
        elif self.start_at is ConsumerStartPosition.LATEST:
            # FIXME: librdkafka has a bug in offset handling - it caches
            # "OFFSET_END", and will repeatedly move to the end of the
            # topic. See https://github.com/edenhill/librdkafka/pull/2876 -
            # it should get fixed in v1.5 of librdkafka.

            librdkafka_version = confluent_kafka.libversion()[0]
            if librdkafka_version < "1.5.0":
                self.logger.warn(
                    "In librdkafka before v1.5, LATEST offsets have buggy behavior; you may "
                    f"not receive data (your librdkafka version is {librdkafka_version}). See "
                    "https://github.com/confluentinc/confluent-kafka-dotnet/issues/1254.")
            default_topic_config = config.get("default.topic.config", {})
            default_topic_config = {
                "auto.offset.reset": "LATEST",
            }
            config["default.topic.config"] = default_topic_config

        if self.auth is not None:
            config.update(self.auth())
        return config



import unittest

from adc import kafka


class TestKafkaURLParsing(unittest.TestCase):
    def test_fully_populated(self):
        group, brokers, topic = kafka.parse_kafka_url(
            "kafka://group@broker/topic")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["broker"])
        self.assertEqual(topic, ["topic"])

    def test_multiple_broker(self):
        group, brokers, topic = kafka.parse_kafka_url(
            "kafka://broker1,broker2/topic")
        self.assertIs(group, None)
        self.assertListEqual(brokers, ["broker1", "broker2"])
        self.assertEqual(topic, ["topic"])

    def test_multiple_broker_with_group(self):
        group, brokers, topic = kafka.parse_kafka_url(
            "kafka://group@broker1,broker2/topic")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["broker1", "broker2"])
        self.assertEqual(topic, ["topic"])

    def test_multiple_topics(self):
        group, brokers, topic = kafka.parse_kafka_url(
            "kafka://group@broker1,broker2/topic1,topic2")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["broker1", "broker2"])
        self.assertEqual(topic, ["topic1", "topic2"])

    def test_hostport_address(self):
        group, brokers, topic = kafka.parse_kafka_url(
            "kafka://group@127.0.0.1:9092/topic")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["127.0.0.1:9092"])
        self.assertEqual(topic, ["topic"])

    def test_no_group(self):
        group, brokers, topic = kafka.parse_kafka_url("kafka://broker/topic")
        self.assertIs(group, None)
        self.assertListEqual(brokers, ["broker"])
        self.assertEqual(topic, ["topic"])

    def test_no_topic(self):
        group, brokers, topic = kafka.parse_kafka_url("kafka://group@broker")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["broker"])
        self.assertIs(topic, None)

    def test_no_topic_trailing_slash(self):
        group, brokers, topic = kafka.parse_kafka_url("kafka://group@broker/")
        self.assertEqual(group, "group")
        self.assertListEqual(brokers, ["broker"])
        self.assertIs(topic, None)

    def test_bad_scheme(self):
        with self.assertRaises(ValueError):
            kafka.parse_kafka_url("http://group@broker")

    def test_no_scheme(self):
        with self.assertRaises(ValueError):
            kafka.parse_kafka_url("group@broker")

    def test_empty_scheme(self):
        with self.assertRaises(ValueError):
            kafka.parse_kafka_url("://group@broker")



from setuptools import setup


# requirements
install_requires = [
    "confluent-kafka",
    "dataclasses ; python_version < '3.7'",
    "importlib-metadata ; python_version < '3.8'",
    "tqdm",
    "certifi>=2020.04.05.1"
]

dev_requires = [
    "autopep8",
    "docker",
    "flake8",
    "isort",
    "pytest",
    "pytest-timeout",
    "pytest-integration",
    "sphinx",
    "sphinx_rtd_theme",
    "twine",
]


setup(name='adc-streaming',
      description='Astronomy Data Commons streaming client libraries',
      long_description=open("README.md").read(),
      long_description_content_type="text/markdown",
      url='https://github.com/astronomy-commons/adc-streaming',
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: BSD License",
          "Development Status :: 3 - Alpha",
          "Operating System :: POSIX :: Linux",
          "Operating System :: MacOS :: MacOS X"
      ],
      author='Astronomy Data Commons Team',
      author_email='swnelson@uw.edu',
      license='BSD',
      packages=['adc'],
      install_requires=install_requires,
      extras_require={
          "dev": dev_requires,
      },
      setup_requires=['setuptools_scm'],
      use_scm_version=True,
      zip_safe=False)
